/*
 * Copyright (C) 2003-2004 Nokia
 * Copyright (C) 2001      BitWagon Software LLC
 * Copyright (C) 1993-2003 Free Software Foundation
 *
 * This file may be used, modified, and distributed under the terms
 * of the GNU General Public License version 2.
 */

#define _GNU_SOURCE

#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/syscall.h>
#include <unistd.h>

#include <fcntl.h>
#include <stdarg.h>
#include <dlfcn.h>
#include <libgen.h>
#include <limits.h>
#include <errno.h>
#include <paths.h>

#include <sys/utsname.h>
#include <dirent.h>
#include <sys/stat.h>

#include <elf.h>
#include <sys/mman.h>

#include <sys/time.h>
#include <time.h>

#include <ctype.h>
#include "sb_env.h"


#ifndef STATIC
# ifdef RTLD_NEXT
#  define USE_RTLD_NEXT
# endif
#endif


/*
 * execve wrapping is taken directly from fakeroot
 * please look at fakeroot source code for examples, documentation
 * and overall understanding how this works :-)
 * */


#define RPATH_PREFIX "/scratchbox/"
#define LDPATH       "/scratchbox/host_shared/lib/ld.so"
#define LIBCPATH     "/scratchbox/host_shared/lib/libc.so.6"

#define BUFFERSIZE PATH_MAX

#ifdef DEBUG
#define DEBUG_OUT(...) fprintf(stderr, __VA_ARGS__)
#else
#define DEBUG_OUT(...)
#endif


extern char **environ;

struct next_wrap_st {
	void **doit;
	char *name;
};

static void *get_libc(void)
{
#ifndef USE_RTLD_NEXT
	static void *lib = 0;
	if (!lib) {
		lib = dlopen(LIBCPATH, RTLD_LAZY);
	}
	if (NULL == lib) {
		fprintf(stderr, "dlopen: %s: %s\n", LIBCPATH, dlerror());
		abort();
	}
	return lib;
#else
	return RTLD_NEXT;
#endif
}

static void load_library_symbols(void);

typedef int (*execve_fun)(const char *, char *const [], char *const []);
typedef int (*uname_fun)(struct utsname *buf);
typedef int (*readlink_fun)(const char *, char *, size_t);

#ifndef STATIC
#define next_execve_real next_execve
#define next_uname_real next_uname
#define next_readlink_real next_readlink
#endif

execve_fun next_execve_real;
uname_fun next_uname_real;
readlink_fun next_readlink_real;

#ifdef STATIC
int *errno_p=NULL;
#define ERRNO_WRAP(x,y,z) static int x y {\
int rc=x##_real z;\
errno=*errno_p;\
return rc;\
}

ERRNO_WRAP(next_execve,(const char *file, char *const argv[], char *const envp[]),(file,argv,envp));
ERRNO_WRAP(next_uname,(struct utsname*buf),(buf));
ERRNO_WRAP(next_readlink,(const char *path, char *buf, size_t bufsiz),(path,buf,bufsiz));

#endif


static int tmp_execve(const char *file, char *const argv[], char *const envp[])
{
	load_library_symbols();
	return next_execve_real(file, argv, envp);
}

static int tmp_uname(struct utsname *buf)
{
	load_library_symbols();
	return next_uname_real(buf);
}

static int tmp_readlink(const char *path, char *buf, size_t bufsiz)
{
	load_library_symbols();
	return next_readlink_real(path, buf, bufsiz);
}


execve_fun next_execve_real=tmp_execve;
uname_fun next_uname_real=tmp_uname;
readlink_fun next_readlink_real=tmp_readlink;


static struct next_wrap_st next_wrap[] = {
	{ (void (*)) &next_execve_real, "execve" },
	{ (void (*)) &next_uname_real,  "uname"  },
	{ (void (*)) &next_readlink_real, "readlink" },
#ifdef STATIC
	{ (void (*)) &errno_p, "errno" },
#endif
	{ NULL, NULL }
};

static void load_library_symbols(void)
{
	static int done = 0;
	int i;

	if (done) {
		return;
	}

	for (i = 0; next_wrap[i].doit; i++) {
		char *msg;
		*(next_wrap[i].doit) = dlsym(get_libc(), next_wrap[i].name);
		if ((msg = dlerror())) {
			fprintf(stderr, "dlsym(%s): %s\n",
				next_wrap[i].name, msg);
		}
	}
}


enum binary_type { BIN_NONE, BIN_UNKNOWN, BIN_FOREIGN,
				   BIN_STATIC, BIN_DYNAMIC, BIN_SCRATCHBOX, BIN_ALIEN };

#define PAGE_MASK (~0u << 12)

#if __BYTE_ORDER == __BIG_ENDIAN
# define elf_endianness ELFDATA2MSB
#elif __BYTE_ORDER == __LITTLE_ENDIAN
# define elf_endianness ELFDATA2LSB
#else
# error Invalid __BYTE_ORDER
#endif

/* TODO: should we drop class testing? */
#define elf_class   ELFCLASS32

/* TODO: how can this be made portable? */
#define elf_machine EM_386

/*
 * Returns binary type that best describes the ELF binary.
 */
static enum binary_type inspect_binary(const char *filename)
{
	enum binary_type retval;
	int fd, phnum, j;
	struct stat status;
	char *region;
	unsigned int reloc0, ph_base, ph_frag;
	Elf32_Ehdr *ehdr;
	Elf32_Phdr *phdr;

	retval = BIN_NONE;

	fd = open(filename, O_RDONLY, 0);
	if (fd < 0) {
		goto _out;
	}

	retval = BIN_UNKNOWN;

	if (fstat(fd, &status) < 0) {
		goto _out_close;
	}

	if (!S_ISREG(status.st_mode) && !S_ISLNK(status.st_mode)) {
		goto _out_close;
	}

	region = mmap(0, status.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
	if (!region) {
		goto _out_close;
	}

	ehdr = (Elf32_Ehdr *) region;

	if (strncmp((char *) ehdr, ELFMAG, SELFMAG) != 0) {
		goto _out_munmap;
	}

	retval = BIN_FOREIGN;

	if (ehdr->e_ident[EI_DATA] != elf_endianness ||
	    ehdr->e_ident[EI_CLASS] != elf_class ||
	    ehdr->e_ident[EI_VERSION] != EV_CURRENT ||
	    ehdr->e_machine != elf_machine) {
		goto _out_munmap;
	}

	retval = BIN_STATIC;

	phnum = ehdr->e_phnum;
	reloc0 = ~0;
	ph_base = ehdr->e_phoff & PAGE_MASK;
	ph_frag = ehdr->e_phoff - ph_base;

	phdr = (Elf32_Phdr *) (region + ph_base + ph_frag);

	for (j = phnum; --j >= 0; ++phdr) {
		if (PT_LOAD == phdr->p_type && ~0 == reloc0) {
			reloc0 = phdr->p_vaddr - phdr->p_offset;
		}
	}

	phdr -= phnum;

	for (j = phnum; --j >= 0; ++phdr) {
		int szDyn;
		unsigned int pt_base, pt_frag;
		Elf32_Dyn *dp_rpath, *dp_strsz, *dp_strtab, *dp;

		if (PT_DYNAMIC != phdr->p_type) {
			continue;
		}

		retval = BIN_DYNAMIC;

		pt_base = phdr->p_offset & PAGE_MASK;
		pt_frag = phdr->p_offset - pt_base;
		szDyn = phdr->p_filesz;

		dp_rpath = NULL;
		dp_strsz = NULL;
		dp_strtab = NULL;

		dp = (Elf32_Dyn *) ((char *) region + pt_base + pt_frag);

		for (; 0 <= (szDyn -= sizeof (*dp)); ++dp) {
			if (DT_RPATH == dp->d_tag) {
				dp_rpath = dp;
			} else if (DT_STRTAB == dp->d_tag) {
				dp_strtab = dp;
			} else if (DT_STRSZ == dp->d_tag) {
				dp_strsz = dp;
			}
		}

		if (dp_rpath && dp_strtab && dp_strsz) {
			char *rpathstr, *strtab, *match;
			unsigned int unloc, str_base, str_frag;

			unloc = dp_strtab->d_un.d_val - reloc0;
			str_base = unloc & PAGE_MASK;
			str_frag = unloc - str_base;

			strtab = (char *) (region + str_base + str_frag);
			rpathstr = dp_rpath->d_un.d_val + strtab;

			match = strstr(rpathstr, RPATH_PREFIX);
			if (match && (match == rpathstr || match[-1] == ':')) {
				retval = BIN_SCRATCHBOX;
				break;
			}
		}
	}
	{
		char resolved[PATH_MAX]="";
		if(BIN_DYNAMIC == retval && realpath(filename,resolved) )
			if(strncmp(resolved,"/scratchbox",strlen("/scratchbox"))==0 ||
			   strncmp(resolved,"/host_usr",strlen("/host_usr"))==0)
				retval=BIN_ALIEN;
	}

_out_munmap:
	munmap(region, status.st_size);
_out_close:
	close(fd);
_out:
	return retval;
}

static char * const *select_preload(char * const oldenv[], int target_binary)
{
	char ** newenv;
	int i, newenv_tail;
	const char * sbox_preload;

	/* Calculate vector length  */

	for (i = 0; oldenv[i]; ++i)
		;

	/* Allocate space for vector, a possibly-new LD_PRELOAD variable
	 * and null-terminator  */

	newenv = (char **) calloc(i + 2, sizeof (char *));
	if (!newenv) {
		fprintf(stderr, "libsb: %s\n", strerror(errno));
		return oldenv;
	}

	/* Copy vector and remove old LD_PRELOAD if any  */

	for (i = 0, newenv_tail = 0; oldenv[i]; ++i) {
		if (strncmp(oldenv[i], "LD_PRELOAD=", 11) != 0) {
			newenv[newenv_tail++] = oldenv[i];
		}
	}

	/* Set correct LD_PRELOAD alternative depending on binary type  */

	sbox_preload = getenv("SBOX_PRELOAD");
	if (sbox_preload) {
		const char * delim;
		size_t len;
		const char * str;
		char * var;

		delim = strchr(sbox_preload, ',');

		if (target_binary) {
			if (delim)
				str = delim + 1;
			else
				str = "";
			len = strlen(str);
		} else {
			str = sbox_preload;
			if (delim)
				len = delim - sbox_preload;
			else
				len = strlen(str);
		}

		var = malloc(strlen("LD_PRELOAD=") + len + 1);
		if (var) {
			strcpy(var, "LD_PRELOAD=");
			strncat(var, str, len);

			newenv[newenv_tail++] = var;
			DEBUG_OUT("Setting LD_PRELOAD to be '%s'\n",var);
		} else
			fprintf(stderr, "libsb: %s\n", strerror(errno));
	}

	return newenv;
}




enum log_type { LOG_NONE, LOG_DIRECTORY, LOG_BINARY, LOG_IGNORE, LOG_SCRIPT, LOG_LIMIT };

static FILE *log_open()
{
	char *filename;
	FILE *file;
	struct timeval tv;
	struct tm *t;

	gettimeofday(&tv, NULL);
	t = gmtime(&tv.tv_sec);

	filename = getenv("SBOX_REDIRECT_LOG");
	if (!filename || filename[0] == '\0') {
		return NULL;
	}

	file = fopen(filename, "a");
	if (!file) {
		fprintf(stderr, "libsb: %s: %s\n", filename, strerror(errno));
		return NULL;
	}

	fprintf(file, "[%04d-%02d-%02d %02d:%02d:%02d %5d] ",
		1900 + t->tm_year, t->tm_mon + 1, t->tm_mday,
		t->tm_hour, t->tm_min, t->tm_sec, getpid());

	return file;
}

static void log_redirect(enum log_type action, const char *from, const char *to, const char *arg,
			 unsigned int recursion, unsigned int script_recursion)
{
	FILE *file;

	file = log_open();
	if (!file)
		return;

	fprintf(file, "Redirect(%d): ", recursion);

	switch (action) {
	case LOG_NONE:
		fprintf(file, "None       %s", from);
		break;

	case LOG_DIRECTORY:
		fprintf(file, "Directory  %s -> %s", from, to);
		break;

	case LOG_BINARY:
		fprintf(file, "Binary     %s -> %s", from, to);
		break;

	case LOG_IGNORE:
		fprintf(file, "Ignored    %s", from);
		break;

	case LOG_SCRIPT:
		if (arg)
			fprintf(file, "Script(%d)  %s -> %s %s %s", script_recursion, from, to, arg, from);
		else
			fprintf(file, "Script(%d)  %s -> %s %s", script_recursion, from, to, from);
		break;

	case LOG_LIMIT:
		fprintf(file, "Limit(%d/%d) %s", recursion, script_recursion, from);
		break;
	}

	fprintf(file, "\n");
	fclose(file);
}

static void log_binary(enum binary_type type)
{
	FILE *file;

	file = log_open();
	if (!file)
		return;

	fprintf(file, "Binary type: ");

	switch (type) {
	case BIN_NONE:
		fprintf(file, "None");
		break;

	case BIN_UNKNOWN:
		fprintf(file, "Unknown");
		break;

	case BIN_FOREIGN:
		fprintf(file, "Foreign    Dropping preloading");
		break;

	case BIN_STATIC:
		fprintf(file, "Static     Dropping preloading; Translating environment");
		break;

	case BIN_DYNAMIC:
		fprintf(file, "Dynamic    Dropping preloading; Translating environment");
		break;

	case BIN_SCRATCHBOX:
		fprintf(file, "Scratchbox");
		break;
	case BIN_ALIEN:
		fprintf(file, "Alien");
		break;
	}

	fprintf(file, "\n");
	fclose(file);
}


/*
 * Checks if a file has #! and reads the interpreter and its argument to
 * buffers of 1024 bytes.
 */
static int check_for_hashbang(const char *filename,
			      char **interp_buf,
			      char **interparg_buf)
{
	int hash_bang = 0;
	FILE *fp;

	if ((fp = fopen(filename, "r"))) {
		char buf[BUFFERSIZE];  /* hash bang detection buffer */

		memset(buf, 0, BUFFERSIZE);
		fread(buf, BUFFERSIZE - 1, 1, fp);

		if (buf[0] == '#' && buf[1] == '!') {
			int begin, end;

			/* find begin of interpreter */

			begin = 2;
			while (buf[begin] == ' ' || buf[begin] == '\t')
				++begin;

			/* find end of interpreter */

			end = begin;
			while (buf[end] != ' '  && buf[end] != '\t' &&
			       buf[end] != '\0' && buf[end] != '\n')
				++end;

			/* get the interpreter */

			*interp_buf = strndup(&buf[begin], end - begin);

			/* find begin of argument */

			while (buf[end] == ' ' || buf[end] == '\t')
				++end;
			begin = end;

			/* find end of argument */

			while (buf[end] != '\n' && buf[end] != '\0' && !isspace(buf[end]))
				++end;

			/* get the interpreter argument */

			if (begin != end)
				*interparg_buf = strndup(&buf[begin], end - begin);
			else
				*interparg_buf = NULL;

			hash_bang = 1;
		}

		fclose(fp);
	}

	return hash_bang;
}

/*
 * Checks if the directory part of bin_path is under an entry in bindir_list.
 */
static int in_bindirs(const char *bindir_list,
		      const char *bin_path)
{
	char *list, *dir, *bindir;

	list = strdup(bindir_list);
	dir = dirname(strdup(bin_path));

	while ((bindir = strsep(&list, ":"))) {
		if (strcmp(bindir, dir) == 0) {
			return 1;
		}
	}

	return 0;
}

/*
 * Concatenates PATH and FILE into BUF and returns 0. If the length of the
 * result string is over PATH_MAX, -1 will be returned and message is printed
 * to stderr.
 */
static int build_path(const char *path,
		      const char *file,
		      char *buf)
{
	if (strlen(path) + strlen(file) + 2 > PATH_MAX) {
		fprintf(stderr, "libsb/execve: Path is too long: %s/%s",
			path, file);
		return -1;
	}

	strcpy(buf, path);
	strcat(buf, "/");
	strcat(buf, file);
	return 0;
}

static char **update_args(char *const *old_argv, const char *bin, const char *opt, const char *script);

static int do_exec(const char *file, char *const *argv, char *const *envp)
{
	enum binary_type type = inspect_binary(file);
	int target = 0, rc = 0;

	switch (type) {
	case BIN_FOREIGN:
	case BIN_STATIC:
	case BIN_DYNAMIC:
		envp = override_sbox_env(envp);
		target = 1;
		break;

	case BIN_ALIEN:
		argv = update_args(argv,file,file,NULL);
		file = "/scratchbox/tools/bin/sb_alien";
		break;
	case BIN_NONE:
	case BIN_UNKNOWN:
	case BIN_SCRATCHBOX:
		break;
	}

	envp = select_preload(envp, target);

	log_binary(type);

#ifdef MEMCHECK
	exit(0);
#endif

	rc = next_execve(file, argv, envp);
	if(rc!=0) {
		DEBUG_OUT("Running %s while..",file);
#ifdef DEBUG
		perror("execve failed");
#endif
	}
	return rc;
}


static char **update_args(char *const *old_argv, const char *bin, const char *opt, const char *script)
{
	int argc, old_argc, offset, index;
	char const* *argv;

	offset = 0;
	if (opt)
		++offset;
	if (script)
		++offset;

	for (old_argc = 0; old_argv[old_argc]; ++old_argc)
		;

	argc = offset + old_argc;
	argv = calloc(sizeof (char *), argc + 1);

	memcpy(&argv[offset], old_argv, sizeof (char *) * old_argc);

	index = 0;
	argv[index++] = bin;
	if (opt)
		argv[index++] = opt;
	if (script)
		argv[index++] = script;

	return (char **)argv;
}

static int redirect_exec(const char *file, char *const argv[], char *const envp[],
			 unsigned int recursion, unsigned int script_recursion)
{
	char *fromdir_list, *todir_list, *redir_list, *ignore_list;
	char *interpreter, *interpreter_arg;
	enum log_type log_action = LOG_NONE;

	fromdir_list = getenv("SBOX_REDIRECT_FROM_DIRS");
	todir_list = getenv("SBOX_REDIRECT_TO_DIRS");
	redir_list = getenv("SBOX_REDIRECT_BINARIES");
	ignore_list = getenv("SBOX_REDIRECT_IGNORE");

	/* redirect only once */

	if (recursion >= 1) {
		log_action = LOG_LIMIT;
		goto no_redirect;
	}

	/* ignore individual binaries */

	if (ignore_list) {
		char *path;

		ignore_list = strdup(ignore_list);

		while ((path = strsep(&ignore_list, ":")))
			if (strcmp(file, path) == 0) {
				log_action = LOG_IGNORE;
				goto no_redirect;
			}
	}

	/* redirect individual binaries */

	if (redir_list) {
		char *paths;

		redir_list = strdup(redir_list);

		while ((paths = strsep(&redir_list, ","))) {
			char *from, *to;

			from = strsep(&paths, ":");
			if (!from)
				continue;

			to = strsep(&paths, ":");
			if (!to)
				continue;

			if (strcmp(from, file) == 0) {
				char **my_argv = update_args(argv, to, NULL, NULL);

				log_redirect(LOG_BINARY, from, to, NULL, recursion, script_recursion);
				return redirect_exec(to, my_argv, envp, recursion + 1, script_recursion);
			}
		}
	}

	/* redirect binaries in directories */

	if (fromdir_list && todir_list && in_bindirs(fromdir_list, file)) {
		char *bindir;

		todir_list = strdup(todir_list);

		while ((bindir = strsep(&todir_list, ":"))) {
			char fullname[PATH_MAX];
			char **my_argv;
			struct stat f;

			if (build_path(bindir, basename(strdup(file)), fullname) < 0)
				continue;

			if (stat(fullname, &f) < 0)
				continue;

			if (!S_ISREG(f.st_mode))
				continue;

			my_argv = update_args(argv, fullname, NULL, NULL);

			log_redirect(LOG_DIRECTORY, file, fullname, NULL, recursion, script_recursion);
			return redirect_exec(fullname, my_argv, envp, recursion + 1, script_recursion);
		}
	}

no_redirect:

	if (check_for_hashbang(file, &interpreter, &interpreter_arg)) {
		char **my_argv;

		if (script_recursion >= 1) {
			log_action = LOG_LIMIT;
			goto default_handler;
		}

		my_argv = update_args(argv, interpreter, interpreter_arg, (char *) file);

		log_redirect(LOG_SCRIPT, file, interpreter, interpreter_arg, recursion, script_recursion);
		return redirect_exec(interpreter, my_argv, envp, 0, script_recursion + 1);
	}

default_handler:

	log_redirect(log_action, file, NULL, NULL, recursion, script_recursion);
	return do_exec(file, argv, envp);
}

int execve(const char *file,
	   char *const argv[],
	   char *const envp[])
{
	DEBUG_OUT("fake execve!\n");

	/* alien tools are executed explicitly with the dynamic linker */
	if (strcmp(file, LDPATH) == 0) {
		log_redirect(LOG_NONE, file, NULL, NULL, 0, 0);
		log_binary(BIN_SCRATCHBOX);
#ifdef MEMCHECK
		exit(0);
#endif
		return next_execve(file, argv, envp);
	}

	if (strstr(file, "/targets/") == file) {
		char *target_file;
		// The file actually lives in /target space 
		// eg. it is not a symlink out from the target.
		if(NULL!=(target_file=realpath(file,malloc(PATH_MAX))) &&
		   strstr(target_file,"/targets/")==target_file) {
			free(target_file);
			DEBUG_OUT("Removing prefix from file %s.\n",file);
			target_file = strchr(file + strlen("/targets/"), '/');
			// Target file has a name and is in subdirectory.
			if (target_file && strchr(target_file+1,'/')!=NULL) {
				file = target_file;
			}
			DEBUG_OUT("Using %s.\n",file);
		}
	}

	return redirect_exec(file, argv, envp, 0, 0);
}


static void set_name(char *str, const char *name)
{
	char *value = getenv(name);
	if (value) {
		strncpy(str, value, _UTSNAME_LENGTH);
	}
}

int uname(struct utsname *buf)
{
	if (next_uname(buf) < 0) {
		return -1;
	}

	set_name(buf->sysname,  "SBOX_UNAME_SYSNAME");
	set_name(buf->nodename, "SBOX_UNAME_NODENAME");
	set_name(buf->release,  "SBOX_UNAME_RELEASE");
	set_name(buf->version,  "SBOX_UNAME_VERSION");
	set_name(buf->machine,  "SBOX_UNAME_MACHINE");

	return 0;
}




/* Substitute buf with the first target binary argument
 * of the process, that is a path to an executable.
 */
static int fake_readlink(const char *exe_path, char *buf, size_t bufsiz)
{
	char *dir_mem, *dir, *path, *choice = NULL;
	char mem[4096];
	int fd, i, len, retval;

	retval = -1;

	dir_mem = strdup(exe_path);
	if (!dir_mem)
		goto no_dir;

	dir = dirname(dir_mem);

	path = malloc(strlen(dir) + strlen("/cmdline") + 1);
	if (!path)
		goto no_path;

	strcpy(path, dir);
	strcat(path, "/cmdline");

	fd = open(path, O_RDONLY);
	if (fd < 0)
		goto no_fd;

	len = read(fd, mem, sizeof (mem));
	if (len <= 0)
		goto no_read;

	choice = 0;

	for (i = 0; i < len - 1; i++)
		if (mem[i] == '\0') {
			char *arg;
			struct stat st;

			arg = &mem[i + 1];

			if (stat(arg, &st) < 0)
				continue;

			if (S_ISDIR(st.st_mode) || !(st.st_mode & S_IXGRP))
				continue;

			if ( inspect_binary( arg ) == BIN_FOREIGN ) {
				choice = arg;
                                break;
                        }
		}

	if (choice) {
		retval = strlen(choice);
		strncpy(buf, choice, retval);
		buf[retval]='\0';
	} else {
		retval = strlen(mem);
		strncpy(buf,mem,strlen(mem));
	}

no_read:
	close(fd);
no_fd:
	free(path);
no_path:
	free(dir_mem);
no_dir:
	return retval;
}

/*
 * readlink needs to be wrapped because sometimes programs want
 * to know their cmdline by inspecting /proc/$PID/exe
 * return value of readlink has to be changed if:
 *
 * 1) environment variable SBOX_READLINK_HACK is set to true
 * 2) inspected symlink is /proc/$PID/exe
 * 3) /proc/$PID/cmdline starts with cpu-transparency method
 *
 * if all these preconditions are true, then return first
 * foreign binary found in /proc/$PID/cmdline
 *
 */
int readlink(const char *path, char *buf, size_t bufsiz)
{
	static const char *location_prefix[] = { "/scratchbox/", "/host_usr/", 0 };
	static const char *binary_prefix[] = { "sbrsh", "qemu", 0 };

	int retval;
	char *env, *base;

	retval = next_readlink(path, buf, bufsiz);
	if (retval < 0)
		return retval;

	env = getenv("SBOX_READLINK_HACK");
	if (!env || strcmp(env,"yes"))
		return retval;

	/* GNU basename lets us do this */
	base = basename((char *) path);

	/* Is path of the form "/proc/.../exe"? */
	if (strstr(path, "/proc/") == path && strcmp(base, "exe") == 0) {
		int i;
		char *target_path, *target_bin;

		target_path = strndup(buf, retval);
		if (!target_path)
			return -1;

		/* Is it in a known host tool directory? */
		for (i = 0; location_prefix[i]; i++)
			if (strstr(target_path, location_prefix[i]) == target_path)
				break;
		if (!location_prefix[i])
			goto no_hack;

		target_bin = basename(target_path);

		/* Is it a known CPU-transparency method? */
		for (i = 0; binary_prefix[i]; i++)
			if (strstr(target_bin, binary_prefix[i]) == target_bin)
				break;
		if (!binary_prefix[i])
			goto no_hack;

		free(target_path);

		return fake_readlink(path, buf, bufsiz);

	no_hack:
		free(target_path);

		return retval;
	}

	return retval;
}


/* ************ exec family of functions copied from glibc-2.3.2 ***********/


#ifndef	HAVE_GNU_LD
# define __environ	environ
#endif

/* Execute PATH with all arguments after PATH until
   a NULL pointer and environment from `environ'.  */
int
execl (const char *path, const char *arg, ...)
{
  size_t argv_max = 1024;
  const char **argv = alloca (argv_max * sizeof (const char *));
  unsigned int i;
  va_list args;

  argv[0] = arg;

  va_start (args, arg);
  i = 0;
  while (argv[i++] != NULL)
    {
      if (i == argv_max)
	{
	  const char **nptr = alloca ((argv_max *= 2) * sizeof (const char *));

#ifndef _STACK_GROWS_UP
	  if ((char *) nptr + argv_max == (char *) argv)
	    {
	      /* Stack grows down.  */
	      argv = (const char **) memcpy (nptr, argv,
					     i * sizeof (const char *));
	      argv_max += i;
	    }
	  else
#endif
#ifndef _STACK_GROWS_DOWN
	    if ((char *) argv + i == (char *) nptr)
	    /* Stack grows up.  */
	    argv_max += i;
	  else
#endif
	    /* We have a hole in the stack.  */
	    argv = (const char **) memcpy (nptr, argv,
					   i * sizeof (const char *));
	}

      argv[i] = va_arg (args, const char *);
    }
  va_end (args);

  return execve (path, (char *const *) argv, __environ);
}




/* Execute PATH with all arguments after PATH until a NULL pointer,
   and the argument after that for environment.  */
int
execle (const char *path, const char *arg, ...)
{
  size_t argv_max = 1024;
  const char **argv = alloca (argv_max * sizeof (const char *));
  const char *const *envp;
  unsigned int i;
  va_list args;
  argv[0] = arg;

  va_start (args, arg);
  i = 0;
  while (argv[i++] != NULL)
    {
      if (i == argv_max)
	{
	  const char **nptr = alloca ((argv_max *= 2) * sizeof (const char *));

#ifndef _STACK_GROWS_UP
	  if ((char *) nptr + argv_max == (char *) argv)
	    {
	      /* Stack grows down.  */
	      argv = (const char **) memcpy (nptr, argv,
					     i * sizeof (const char *));
	      argv_max += i;
	    }
	  else
#endif
#ifndef _STACK_GROWS_DOWN
	    if ((char *) argv + i == (char *) nptr)
	    /* Stack grows up.  */
	    argv_max += i;
	  else
#endif
	    /* We have a hole in the stack.  */
	    argv = (const char **) memcpy (nptr, argv,
					   i * sizeof (const char *));
	}

      argv[i] = va_arg (args, const char *);
    }

  envp = va_arg (args, const char *const *);
  va_end (args);

  return execve (path, (char *const *) argv, (char *const *) envp);
}




/* Execute FILE, searching in the `PATH' environment variable if
   it contains no slashes, with all arguments after FILE until a
   NULL pointer and environment from `environ'.  */
int
execlp (const char *file, const char *arg, ...)
{
  size_t argv_max = 1024;
  const char **argv = alloca (argv_max * sizeof (const char *));
  unsigned int i;
  va_list args;

  argv[0] = arg;

  va_start (args, arg);
  i = 0;
  while (argv[i++] != NULL)
    {
      if (i == argv_max)
	{
	  const char **nptr = alloca ((argv_max *= 2) * sizeof (const char *));

#ifndef _STACK_GROWS_UP
	  if ((char *) nptr + argv_max == (char *) argv)
	    {
	      /* Stack grows down.  */
	      argv = (const char **) memcpy (nptr, argv,
					     i * sizeof (const char *));
	      argv_max += i;
	    }
	  else
#endif
#ifndef _STACK_GROWS_DOWN
	    if ((char *) argv + i == (char *) nptr)
	    /* Stack grows up.  */
	    argv_max += i;
	  else
#endif
	    /* We have a hole in the stack.  */
	    argv = (const char **) memcpy (nptr, argv,
					   i * sizeof (const char *));
	}

      argv[i] = va_arg (args, const char *);
    }
  va_end (args);

  return execvp (file, (char *const *) argv);
}


#ifndef	HAVE_GNU_LD
# define __environ	environ
#endif

/* Execute PATH with arguments ARGV and environment from `environ'.  */
int
execv (const char *path, char *const argv[])
{
  return execve (path, argv, __environ);
}




/* The file is accessible but it is not an executable file.  Invoke
   the shell to interpret it as a script.  */
static void
sb_script_execute (const char *file, char *const argv[])
{
  /* Count the arguments.  */
  int argc = 0;
  while (argv[argc++])
    ;

  /* Construct an argument list for the shell.  */
  {
    char *new_argv[argc + 1];
    new_argv[0] = (char *) _PATH_BSHELL;
    new_argv[1] = (char *) file;
    while (argc > 1)
      {
	new_argv[argc] = argv[argc - 1];
	--argc;
      }

    /* Execute the shell.  */
    execve (new_argv[0], new_argv, __environ);
  }
}


/* Execute FILE, searching in the `PATH' environment variable if it contains
   no slashes, with arguments ARGV and environment from `environ'.  */
int
execvp (file, argv)
     const char *file;
     char *const argv[];
{
  if (*file == '\0')
    {
      /* We check the simple case first. */
      errno = ENOENT;
      return -1;
    }

  if (strchr (file, '/') != NULL)
    {
      /* Don't search when it contains a slash.  */
      execve (file, argv, __environ);

      if (errno == ENOEXEC)
	sb_script_execute (file, argv);
    }
  else
    {
      int got_eacces = 0;
      char *path, *p, *name;
      size_t len;
      size_t pathlen;

      path = getenv ("PATH");
      if (path == NULL)
	{
	  /* There is no `PATH' in the environment.
	     The default search path is the current directory
	     followed by the path `confstr' returns for `_CS_PATH'.  */
	  len = confstr (_CS_PATH, (char *) NULL, 0);
	  path = (char *) alloca (1 + len);
	  path[0] = ':';
	  (void) confstr (_CS_PATH, path + 1, len);
	}

      len = strlen (file) + 1;
      pathlen = strlen (path);
      name = alloca (pathlen + len + 1);
      /* Copy the file name at the top.  */
      name = (char *) memcpy (name + pathlen + 1, file, len);
      /* And add the slash.  */
      *--name = '/';

      p = path;
      do
	{
	  char *startp;

	  path = p;
	  p = strchrnul (path, ':');

	  if (p == path)
	    /* Two adjacent colons, or a colon at the beginning or the end
	       of `PATH' means to search the current directory.  */
	    startp = name + 1;
	  else
	    startp = (char *) memcpy (name - (p - path), path, p - path);

	  /* Try to execute this name.  If it works, execv will not return.  */
	  execve (startp, argv, __environ);

	  if (errno == ENOEXEC)
	    sb_script_execute (startp, argv);

	  switch (errno)
	    {
	    case EACCES:
	      /* Record the we got a `Permission denied' error.  If we end
		 up finding no executable we can use, we want to diagnose
		 that we did find one but were denied access.  */
	      got_eacces = 1;
	    case ENOENT:
	    case ESTALE:
	    case ENOTDIR:
	      /* Those errors indicate the file is missing or not executable
		 by us, in which case we want to just try the next path
		 directory.  */
	      break;

	    default:
	      /* Some other error means we found an executable file, but
		 something went wrong executing it; return the error to our
		 caller.  */
	      return -1;
	    }
	}
      while (*p++ != '\0');

      /* We tried every element and none of them worked.  */
      if (got_eacces)
	/* At least one failure was due to permissions, so report that
           error.  */
	errno = EACCES;
    }

  /* Return the error from the last attempt (probably ENOENT).  */
  return -1;
}


#include	<sys/types.h>
#include	<sys/wait.h>
#include	<errno.h>
#include	<unistd.h>

int
system(const char *cmdstring)	/* version without signal handling */
{
	pid_t	pid;
	int		status;

	if (cmdstring == NULL)
		return(1);		/* always a command processor with Unix */

	if ( (pid = fork()) < 0) {
		status = -1;	/* probably out of processes */

	} else if (pid == 0) {				/* child */
		execl("/bin/sh", "sh", "-c", cmdstring, (char *) 0);
		_exit(127);		/* execl error */

	} else {							/* parent */
		while (waitpid(pid, &status, 0) < 0)
			if (errno != EINTR) {
				status = -1; /* error other than EINTR from waitpid() */
				break;
			}
	}

	return(status);
}

#include	<sys/wait.h>
#include	<errno.h>
#include	<fcntl.h>

static pid_t	*childpid = NULL;
						/* ptr to array allocated at run-time */
static int		maxfd;	/* from our open_max(), {Prog openmax} */

#define	SHELL	"/scratchbox/tools/bin/bash"

FILE *
popen(const char *cmdstring, const char *type)
{
	int		i, pfd[2];
	pid_t	pid;
	FILE	*fp;

			/* only allow "r" or "w" */
	if ((type[0] != 'r' && type[0] != 'w') || type[1] != 0) {
		errno = EINVAL;		/* required by POSIX.2 */
		return(NULL);
	}

	if (childpid == NULL) {		/* first time through */
				/* allocate zeroed out array for child pids */
		maxfd = FOPEN_MAX;
		if ( (childpid = calloc(maxfd, sizeof(pid_t))) == NULL)
			return(NULL);
	}

	if (pipe(pfd) < 0)
		return(NULL);	/* errno set by pipe() */

	if ( (pid = fork()) < 0)
		return(NULL);	/* errno set by fork() */
	else if (pid == 0) {							/* child */
		if (*type == 'r') {
			close(pfd[0]);
			if (pfd[1] != STDOUT_FILENO) {
				dup2(pfd[1], STDOUT_FILENO);
				close(pfd[1]);
			}
		} else {
			close(pfd[1]);
			if (pfd[0] != STDIN_FILENO) {
				dup2(pfd[0], STDIN_FILENO);
				close(pfd[0]);
			}
		}
			/* close all descriptors in childpid[] */
		for (i = 0; i < maxfd; i++)
			if (childpid[i] > 0)
				close(i);

		execl(SHELL, "sh", "-c", cmdstring, (char *) 0);
		_exit(127);
	}
								/* parent */
	if (*type == 'r') {
		close(pfd[1]);
		if ( (fp = fdopen(pfd[0], type)) == NULL)
			return(NULL);
	} else {
		close(pfd[0]);
		if ( (fp = fdopen(pfd[1], type)) == NULL)
			return(NULL);
	}
	childpid[fileno(fp)] = pid;	/* remember child pid for this fd */
	return(fp);
}

int
pclose(FILE *fp)
{
	int		fd, stat;
	pid_t	pid;

	if (childpid == NULL)
		return(-1);		/* popen() has never been called */

	fd = fileno(fp);
	if ( (pid = childpid[fd]) == 0)
		return(-1);		/* fp wasn't opened by popen() */

	childpid[fd] = 0;
	if (fclose(fp) == EOF)
		return(-1);

	while (waitpid(pid, &stat, 0) < 0)
		if (errno != EINTR)
			return(-1);	/* error other than EINTR from waitpid() */

	return(stat);	/* return child's termination status */
}
